% ALU instructions
%
% - thread-wise module: operations described for a single thread
%
% - reference to shuffle operations:
% http://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#warp-description
% https://devblogs.nvidia.com/parallelforall/faster-parallel-reductions-kepler/
%
:- module(alu, [

	% alu: simulates the execution of alu operations
	% alu(Tid, LocalMemBank, SharedMem, CSMask,
	%	Instr, NewLocalMemBank, NewSharedMem)
	alu/7

]).

:- use_module(isa).
:- use_module(memory).


% addt: V1 <- V2 + V3
alu(Tid, LMB, SM, _, addt(V1, V2, V3), NewLMB, SM) :-
	getval(Tid, LMB, V2, N2),
	getval(Tid, LMB, V3, N3),
	N1 is N2 + N3,
	setval(Tid, LMB, V1, N1, NewLMB).

% atomic increment: V1 <- mem[Vx] + 1, mem[Vx] <- mem[Vx] + 1
alu(Tid, LMB, SM, _, atominc(V1, Vx), NewLMB, NewSM) :-
	getval(Tid, LMB, Vx, Addr),
	getval(SM, Addr, N),
	Ninc is N + 1,
	setval(SM, Addr, Ninc, NewSM),
	setval(Tid, LMB, V1, Ninc, NewLMB).

% load immediate: V <- N
alu(Tid, LMB, SM, _, const(V, N), NewLMB, SM) :-
	setval(Tid, LMB, V, N, NewLMB).

% shared memory load: V1 <- mem[Vx]
alu(Tid, LMB, SM, _, load(V1, Vx), NewLMB, SM) :-
	getval(Tid, LMB, Vx, Addr),
	getval(SM, Addr, N),
	setval(Tid, LMB, V1, N, NewLMB).

% lth: V1 <- V2 < V3
alu(Tid, LMB, SM, _, lth(V1, V2, V3), NewLMB, SM) :-
	getval(Tid, LMB, V2, N2),
	getval(Tid, LMB, V3, N3),
	N2 < N3,
	setval(Tid, LMB, V1, 1, NewLMB).

alu(Tid, LMB, SM, _, lth(V1, V2, V3), NewLMB, SM) :-
	getval(Tid, LMB, V2, N2),
	getval(Tid, LMB, V3, N3),
	N2 >= N3,
	setval(Tid, LMB, V1, 0, NewLMB).

% shared memory store: mem[Vx] <- V1
alu(Tid, LMB, SM, _, store(V1, Vx), LMB, NewSM) :-
	getval(Tid, LMB, Vx, Addr),
	getval(Tid, LMB, V1, N),
	setval(SM, Addr, N, NewSM).

% direct copy from indexed lane: V1[Tid] <- V1[Vx[Tid]]
alu(Tid, LMB, SM, CSMask, shfl(V1, Vx), NewLMB, SM) :-
	getval(Tid, LMB, Vx, LaneId),
	memberchk(LaneId, CSMask),
	getval(LaneId, LMB, V1, Value),
	setval(Tid, LMB, V1, Value, NewLMB).

alu(Tid, LMB, SM, CSMask, shfl(V1, Vx), NewLMB, SM) :-
	getval(Tid, LMB, Vx, LaneId),
	\+ memberchk(LaneId, CSMask),
	setval(Tid, LMB, V1, _, NewLMB).

% shfl_up: copy from a lane with lower ID relative to caller
% TODO

% shfl_down: copy from a lane with higher ID relative to caller
% TODO

% shfl_xor: copy from a lane based on bitwise XOR of own lane ID
% TODO
