% Control flow instructions
:- module(control, [

	% Simulates the control flow upon an input machine and program
	%
	% simulate(Threads, DivStack, ConStack, ShdMem,
	% 	Program, PC, NewThreads, NewShdMem)
	% - Threads      := threads           [ThreadID]
	% - LocalMemBank := local memory bank [[(variable, value)]]
	% - ShdMem       := shared memory     [(idx, value)]
	% - DivStack     := divergence stack  [(PC, ActiveTs, Label, InactiveTs)]
	% - ConStack     := context stack     (FullMask, [(Ts, DivStack, PC)])
	% - Program      := program           [instruction]
	% - PC           := program counter   integer
	% - NewThreads   := updated threads   [ThreadID]
	% - NewLocalMem  := local memory      [[(variable, value)]]
	% - NewShdMem    := updated shd mem   [(idx, value)]
	simulate/10

]).

:- use_module(alu).
:- use_module(debug).
:- use_module(memory).
:- use_module(divergence).


% halt execution
simulate(T, LMB, SM, [], CS, P, PC, T, LMB, SM) :-
	lookup(P, PC, stop),
	debug(PC, 'stop', T, LMB, [], CS, SM).

simulate(T, LMB, SM, [_|_], _, P, PC, T, LMB, SM) :-
	lookup(P, PC, stop),
	error(PC, stop, 'divergence stack is not empty.').

% conditional jump: jumps to label L1
simulate(T, LMB, SM, ST, CS, P, PC, TT, LMBB, SMM) :-
	lookup(P, PC, branch(V, L)),
	divergeThreads(T, LMB, V, [], T),
	updateDivStack(ST, [], PC, L, STT),
	debug(PC, 'branch true', T, LMB, STT, CS, SM),
	simulate(T, LMB, SM, STT, CS, P, L, TT, LMBB, SMM).

simulate(T, LMB, SM, ST, CS, P, PC, TT, LMBB, SMM) :-
	lookup(P, PC, branch(V, L)),
	divergeThreads(T, LMB, V, T, []),
	updateDivStack(ST, [], PC, L, STT),
	PCC is PC + 1,
	debug(PC, 'branch false', T, LMB, STT, CS, SM),
	simulate(T, LMB, SM, STT, CS, P, PCC, TT, LMBB, SMM).

simulate(T, LMB, SM, ST, CS, P, PC, TT, LMBB, SMM) :-
	Instr = branch(V, L),
	lookup(P, PC, Instr),
	divergeThreads(T, LMB, V, [Tz|ZT], [Tn|NT]),
	updateDivStack(ST, [Tn|NT], PC, L, STT),
	PCC is PC + 1,
	debug(PC, Instr, [Tz|ZT], LMB, STT, CS, SM),
	simulate([Tz|ZT], LMB, SM, STT, CS, P, PCC, TT, LMBB, SMM).

% conditional jump: jumps to label L1 if Thread is active on CSMask
simulate(T, LMB, SM, ST, CS, P, PC, TT, LMBB, SMM) :-
	Instr = branch_mask(Thread, L),
	lookup(P, PC, Instr),
	(_, [(CSMask, _, _)|_]) = CS,
	memberchk(Thread, CSMask),
	debug(PC, Instr, T, LMB, ST, CS, SM),
	simulate(T, LMB, SM, ST, CS, P, L, TT, LMBB, SMM).

simulate(T, LMB, SM, ST, CS, P, PC, TT, LMBB, SMM) :-
	lookup(P, PC, branch_mask(Thread, L)),
	(_, [(CSMask, _, _)|_]) = CS,
	\+ memberchk(Thread, CSMask),
	PCC is PC + 1,
	debug(PC, branch_mask(Thread, L), T, LMB, ST, CS, SM),
	simulate(T, LMB, SM, ST, CS, P, PCC, TT, LMBB, SMM).

simulate(T, LMB, SM, _, (_, []), P, PC, T, LMB, SM) :-
	lookup(P, PC, branch_mask(Thread, L)),
	error(PC, branch_mask(Thread, L), 'context stack is empty').

% everywhere block: assures all threads within a warp are active
simulate(T, LMB, SM, ST, (FullWarp, CS), P, PC, NewT, NewLMB, NewSM) :-
	lookup(P, PC, everywhere),
	NewCS = (FullWarp, [(T, ST, PC)|CS]),
	PCC is PC + 1,
	debug(PC, 'everywhere', FullWarp, LMB, [], NewCS, SM),
	simulate(FullWarp, LMB, SM, [], NewCS, P, PCC, NewT, NewLMB, NewSM).

simulate(T, LMB, SM, ST, _, P, PC, T, LMB, SM) :-
	lookup(P, PC, end_everywhere),
	ST \= [],
	error(PC, end_everywhere, 'divergence stack is not empty').

simulate(_, LMB, SM, [], (FullMask, [TopCS|CS]), P, PC, NewT, NewLMB, NewSM) :-
	lookup(P, PC, end_everywhere),
	(T, ST, _) = TopCS,
	PCC is PC + 1,
	debug(PC, 'end everywhere', T, LMB, ST, (FullMask, CS), SM),
	simulate(T, LMB, SM, ST, (FullMask, CS), P, PCC, NewT, NewLMB, NewSM).

% synchronization barrier: synch threads within warp
simulate(ZT, LMB, SM, [(PNT, [], L, [Tn|NT])|ST], CS, P, PC, TT, LMBB, SMM) :-
	lookup(P, PC, sync),
	debug(PC, 'sync start', [Tn|NT], LMB, [(PNT, ZT, L, [])|ST], CS, SM),
	simulate([Tn|NT], LMB, SM, [(PNT, ZT, L, [])|ST], CS, P, L, TT, LMBB, SMM).

simulate(NT, LMB, SM, [(_, ZT, _, [])|ST], CS, P, PC, TTT, LMBB, SMM) :-
	lookup(P, PC, sync),
	append(ZT, NT, TT),
	PCC is PC + 1,
	debug(PC, 'sync stop', TT, LMB, ST, CS, SM),
	simulate(TT, LMB, SM, ST, CS, P, PCC, TTT, LMBB, SMM).

% unconditional jump: jumps to label L1
simulate(T, LMB, SM, ST, CS, P, PC, TT, LMBB, SMM) :-
	lookup(P, PC, jump(L)),
	debug(PC, jump(L), T, LMB, ST, CS, SM),
	simulate(T, LMB, SM, ST, CS, P, L, TT, LMBB, SMM).

% ALU instructions
simulate(T, LMB, SM, ST, (FullMask, CS), P, PC, TT, LMBBB, SMMM) :-

	lookup(P, PC, Instr),
	Instr \= (branch(_, _)),
	Instr \= (branch_mask(_, _)),
	Instr \= stop,
	Instr \= everywhere,
	Instr \= end_everywhere,
	Instr \= sync,
	Instr \= (jump(_)),

	(
		[(CSMask, _, _)|_] = CS
	->	simulate_alu(T, LMB, SM, CSMask, Instr, LMBB, SMM)
	;	simulate_alu(T, LMB, SM, FullMask, Instr, LMBB, SMM)
	),

	PCC is PC + 1,
	debug(PC, Instr, T, LMBB, ST, (FullMask, CS), SMM),
	simulate(T, LMBB, SMM, ST, (FullMask, CS), P, PCC, TT, LMBBB, SMMM).

simulate_alu([], LMB, SM, _, _, LMB, SM).
simulate_alu([Tid|Ts], LMB, SM, CSMask, Instr, LMBBB, SMMM) :-
	alu(Tid, LMB, SM, CSMask, Instr, LMBB, SMM),
	simulate_alu(Ts, LMBB, SMM, CSMask, Instr, LMBBB, SMMM).
