% Instruction Set
:- module(isa, [

	% -------------------------------------------------------------------------
	% Arithmetic

	% addt: V3 <- V1 + V2
	addt/3,

	% atomic increment: V2 <- V1 + 1
	atominc/2,

	% load immediate:  V1 <- N
	const/2,

	% shared memory load: V1 <- mem[Addr]
	load/2,

	% lth: V3 <- V1 < V2
	lth/3,

	% shared memory store: mem[Addr] <- V1
	store/2,

	% -------------------------------------------------------------------------
	% Logic

	% conditional jump: jumps to label L1
	branch/2,

	% conditional jump: jumps to label L1 if Thread is active on CSMask
	branch_mask/2,

	% everywhere: begins an everywhere block
	everywhere/0,

	% end_everywhere: ends an everywhere block
	end_everywhere/0,

	% unconditional jump: jumps to label L1
	jump/1,

	% function call
	kernel/1,

	% halt execution
	stop/0

]).


% Maps variables into registers
register(tid).
register(R) :- integer(R).


% -----------------------------------------------------------------------------
% ALU

% addt: V3 <- V1 + V2
addt(V1, V2, V3) :- register(V1), register(V2), register(V3).

% atomic increment: V2 <- V1 + 1
atominc(V1, V2) :- register(V1), register(V2).

% load immediate:  V1 <- N
const(V1, N) :- register(V1), num(N).

% shared memory load: V1 <- mem[Addr]
load(V1, Addr) :- name(V1), name(Addr).

% lth: V3 <- V1 < V2
lth(V1, V2, V3) :- name(V1), name(V2), name(V3).

% shared memory store: mem[Addr] <- V1
store(V1, Addr) :- name(V1), name(Addr).

% direct copy from indexed lane: V1[Tid] <- V1[Vx[Tid]]
shfl(V1, Vx) :- name(V1), name(Vx).

% -----------------------------------------------------------------------------
% Control

% conditional jump: jumps to label L1
branch(V1, L1) :- register(V1), integer(L1).

% conditional jump: jumps to label L1 if Thread is active on CSMask
branch_mask(Thread, L1) :- integer(Thread), integer(L1).

% everywhere: begins an everywhere block
everywhere.

% end_everywhere: ends an everywhere block
end_everywhere.

% unconditional jump: jumps to label L1
jump(L1) :- integer(L1).

% kernel call: performs a kernel call
kernel([_|_]).

% halt execution
stop.
